# This file is part of PyFreeFEM.
#
# PyFreeFEM is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# PyFreeFEM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is included below.
# For further information, see <http://www.gnu.org/licenses/>.
"""PyFreeFEM input/output functions."""
    

import scipy.sparse as sp
import numpy as np
import time
import sys
import subprocess
import threading
import queue
    

try:
    import colored as col

    def colored(text, color=None, attr=None):
        if color:
            text = col.stylize(text, col.fg(color))
        if attr:
            text = col.stylize(text, col.attr(attr))
        return text
except:
    def colored(text, color=None, attr=None):
        return text


def display(message, level=0, debug=0, color=None, attr=None, end='\n',
            flag=None):
    """A function for displaying messages with a flexible tunable level of verbosity.


    :param message:  Text to be printed.
    :param level:   Level of importance of the message which will be actually 
                       printed if ``debug >= level`` or if ``flag`` is ``stdout`` or ``stderr``.
    :type level: int
    :param debug: Desired level of verbosity; the higher and the more information will  
                  be displayed.
    :type debug: int
    :param color: Color of the displayed message. Follows the convention of the     
                  `colored <https://pypi.org/project/colored/>`_ package.
    :param attr:  Formatting of the displayed message. Follows the  
                  convention of the     
                  `colored <https://pypi.org/project/colored/>`_ package.

    :param end:   character appended to the end of a message. Will remove the   
                  final line carriage return if ``end=''``.  

    :param flag:  an indicator to use to indicate if the argument ``message`` comes     
                  from ``stdout`` or ``stderr`` pipe. 
    
    .. note::   
        
       For advanced projects, it can be useful to override :func:`pyfreefem.io.display`,  
       for instance to redirect messages to a file.
    """
    if color or attr:
        message = colored(message, color, attr)
    if debug >= level or flag in ['stdout','stderr']:
        print(message, end=end, flush=True)

tclock = dict()
def tic(ref=0):
    global tclock
    tclock[ref] = time.time()

def toc(ref=0):
    global tclock
    return format(time.time()-tclock[ref],"0.2f")+"s"

class ExecException(Exception):
    """ An exception raised when a FreeFem++ subprocess fails"""
    def __init__(self, message, returncode, stdout, stderr, mix):
        super().__init__(message)
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.mix = mix

def readFFArray(ffarray : str):
    """
    Read an FreeFEM array stored in a file ``ffarray``. 
        
    For instance, if the following ``.edp`` file is executed:

    .. code:: freefem

       real[int] table = [1, 2, 3, 4, 5];
       {
           ifstream f("file.gp");
           f << table;
       }

    Then
        
    .. code:: python

       readFFArray("file.gp") 

    returns the numpy array [1, 2, 3, 4, 5].    
        
    :param ffarray: file name of a FreeFEM array stored with ``ifstream``.
    :return: numpy array identical to ``ffarray``.
    """
    with open(ffarray, "r") as f:
        return np.asarray([float(x) for line in f.readlines()[1:]
                           for x in line.split()])


def writeFFArray(A: np.array, fileName: str):
    """
    Store a numpy array into a FreeFEM array file that can be read with ``ofstream``.

    :param A:  a `numpy.array <https://numpy.org/doc/stable/reference/generated/numpy.array.html>`_ data structure
    :param fileName: an output file name to save the array ``A``.   
    """
    text = f"{len(A)}\t\n"
    lines = [A[5*k:5*(k+1)] for k in range(0, int(np.ceil(len(A)/5)))]
    for line in lines:
        text += ''.join([f"\t  {x}" for x in line])+"\n"
    text = text[:-1]+"\t"
    with open(fileName, "w") as f:
        f.write(text)


def writeFFMatrix(A, fileName):
    """
    Save a dense or a sparse matrix into a FreeFEM matrix file that can be read with ``ofstream``. 

    :param A: A `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_  
              or a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ dense matrix.
    :param fileName: output file name for saving the matrix ``A``.  
    """
    if isinstance(A,np.ndarray):
        preamble = " ".join(map(str,A.shape))+"\t\n"
        preamble += "\n".join(["\t   "+"   ".join(map(str,line)) for line in A.tolist()])
        preamble += "\n\t"
    else:
        (I, J, V) = sp.find(A)
        (n, m) = A.shape
        nnz = A.nnz
        preamble = f"#  HashMatrix Matrix (COO) 0x2d90200\n"
        preamble += "#    n       m        nnz     half     fortran   state  \n"
        preamble += f"{n} {m} {nnz} 0 0 0 0 \n"
        lines = [f"     {i}     {j} {d}" for (i, j, d) in zip(I, J, V)]
        preamble = preamble+"\n".join(lines)
        preamble += "\n"
    with open(fileName, "w") as f:
        f.write(preamble)


def readFFMatrix(ffmatrix : str):
    """
    Read a matrix file generated by a FreeFEM script. For instance,     
    if the following ``.edp`` file is run:  
        
    .. code:: freefem

       matrix table = [[1, 2, 3, 4, 5],[1, 2, 3, 4, 5]];
       {
           ifstream f("file.gp");
           f << table;
       }
    
    Then
        
    .. code:: python

       readFFMatrix("file.gp") 

    returns the numpy matrix [[1, 2, 3, 4, 5],[1, 2, 3, 4, 5]]. 
        
    If the matrix stored by the FreeFEM script is a sparse matrix, then :func:`pyfreefem.io.readFFMatrix`   
    returns a `scipy.sparse.csc_matrix <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.html>`_.

    If it is a dense matrix, then :func:`pyfreefem.io.readFFMatrix` returns a `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        
    :param ffmatrix: file name of a FreeFEM matrix stored with ``ifstream``.
    :return: A `numpy.ndarray` matrix if `ffmatrix` contains a dense matrix, or a `scipy.sparse.csc_matrix` if  
             `ffmatrix` contains a sparse matrix.

    """
    with open(ffmatrix, 'r') as f:
        lines = f.readlines()
    i = []
    j = []
    data = []
    if lines[0].startswith('#  HashMatrix'):
        shape = tuple([int(x) for x in lines[2].strip().split()[:2]])
        if len(lines)>3:
            I,J,data = list(zip(*[line.strip().split() for line in lines[3:]]))
            I = list(map(int,I))
            J = list(map(int,J))
            data=list(map(float,data))
        else:
            I, J= [], []
        return sp.csc_matrix((data, (I,J)),shape=shape)
    elif len(lines[0].strip().split())==2:
        shape = tuple([int(x) for x in lines[0].strip().split()])
        data = [line.strip().split() for line in lines[1:shape[0]+1]]
        return np.asarray(data, dtype=float)
    else:
        for (k, line) in enumerate(lines[4:]):
            i.append(int(line.split()[0])-1)
            j.append(int(line.split()[1])-1)
            data.append(float(line.split()[2]))
    return sp.csc_matrix((data, (i, j)))

def enqueue_stream(stream, queue, type):
    for line in iter(stream.readline, b''):
        queue.put(str(type) + line.decode('utf-8', errors='replace'))
    stream.close()


def enqueue_process(process, queue):
    process.wait()
    queue.put('x')

    
def exec2(cmd, debug=0, level=1, silent=True):
    """Interface with `subprocess.Popen <https://docs.python.org/fr/3/library/subprocess.html>`_.   
       Execute a shell command ``cmd`` and pass standard output and standard errors 
       to the function :func:`pyfreefem.io.display`.
        
       :param cmd: A shell command. For instance ``cmd="FreeFem++ script.edp -v 1 -wg"``.
       :type cmd: str
        
       :param debug: An input ``debug`` parameter, tuning the desired level of verbosity.   
                     The command ``cmd`` will be displayed if ``debug>=level``
       :type debug: int

       :param level: Degree of importance of the executed ``command``.
                     The command ``cmd`` will be displayed if ``debug>=level``
       :type level: int
        
       :param silent: If set to ``True``, no standard output will be displayed.
       :type silent: boolean
        
       .. note::    
            
          By default, the standard output is displayed whatever the value of level and  
          debug because :func:`pyfreefem.io.display` prints ``stdout`` and ``stderr``   
          by watching its ``flag`` argument. Use a custom :func:`pyfreefem.io.display`    
          function to change this behaviour.

    """
    display(colored(cmd, color="indian_red_1a"), level=level, debug=debug, end='',
            flag='shell')
    # Put a line break to separate from the stdout 
    if not silent:
        display("",level=level,debug=debug,flag='')
    tic(121)
    proc = subprocess.Popen("stdbuf -oL "+cmd,shell=True,
                            stdout = subprocess.PIPE,
                            stderr = subprocess.PIPE
                            )
    q = queue.Queue()
    to = threading.Thread(target=enqueue_stream, args=(proc.stdout, q,1))
    te = threading.Thread(target=enqueue_stream, args=(proc.stderr, q,2))
    tp = threading.Thread(target=enqueue_process, args=(proc, q))
    te.start()
    to.start()
    tp.start()

    stdout = "" 
    stderr = "" 
    mix = "" 
    while True:
        line = q.get()
        if line[0] == 'x':
            break
        if line[0] == '1':
            line = line[1:]
            stdout += line
            mix += line 
            if not silent:
                display(line,level=level+1,debug=debug,end='',flag='stdout')
        if line[0] == '2':
            line = line[1:]
            stderr += line
            mix += line 
            if not silent:
                display(line,level=level+1,debug=debug,attr="dim",end='',flag='stderr')
    tp.join()
    te.join()
    to.join()
    if mix and not silent:
        if not mix.endswith('\n'):
            display("",level,debug)
        display("Finished in",level,debug,color="cyan",end="",flag="")
    display(' ('+toc(121)+')', level=level, debug=debug,color="cyan",
            flag="time")

    if proc.returncode != 0:
        raise ExecException('Error : the process "'
                        + colored(cmd, "red")
                        + '" failed with return code '+str(proc.returncode)
                        + ".",proc.returncode,stdout,stderr,mix)
    return proc.returncode, stdout, stderr, mix
    
